'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.sql = exports.createPool = exports.createConnection = exports.transaction = exports.anyFirst = exports.any = exports.manyFirst = exports.many = exports.maybeOneFirst = exports.oneFirst = exports.maybeOne = exports.one = exports.query = exports.NotFoundError = exports.UniqueViolationError = exports.DataIntegrityError = undefined;

var _pg = require('pg');

var _pg2 = _interopRequireDefault(_pg);

var _pgConnectionString = require('pg-connection-string');

var _prettyHrtime = require('pretty-hrtime');

var _prettyHrtime2 = _interopRequireDefault(_prettyHrtime);

var _errors = require('./errors');

var _utilities = require('./utilities');

var _Logger = require('./Logger');

var _Logger2 = _interopRequireDefault(_Logger);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

exports.DataIntegrityError = _errors.DataIntegrityError;
exports.UniqueViolationError = _errors.UniqueViolationError;
exports.NotFoundError = _errors.NotFoundError;


_pg.types.setTypeParser(20, value => {
  return parseInt(value, 10);
});

const log = _Logger2.default.child({
  namespace: 'mightyql'
});

let globalQueryId = 0;

const query = exports.query = async (connection, rawSql, values) => {
  const strippedSql = (0, _utilities.stripComments)(rawSql);

  const queryId = globalQueryId++;

  log.debug({
    queryId,
    sql: strippedSql
  }, 'input query');

  if (values) {
    log.trace({
      queryId,
      values
    }, 'query variables');
  }

  try {
    const start = process.hrtime();

    let result;

    if (Array.isArray(values)) {
      const {
        sql: normalizedSql,
        values: normalizedValues
      } = (0, _utilities.normalizeAnonymousValuePlaceholders)(strippedSql, values);

      result = await connection.query(normalizedSql, normalizedValues);
    } else if (values) {
      const {
        sql: normalizedSql,
        values: normalizedValues
      } = (0, _utilities.normalizeNamedValuePlaceholders)(strippedSql, values);

      result = await connection.query(normalizedSql, normalizedValues);
    } else {
      result = await connection.query(strippedSql);
    }

    const end = process.hrtime(start);

    let rowCount = null;

    if (result.rowCount) {
      rowCount = result.rowCount;
    } else if (Array.isArray(result)) {
      rowCount = result.length;
    }

    log.trace({
      executionTime: (0, _prettyHrtime2.default)(end),
      queryId,
      rowCount
    }, 'query completed');

    return result;
  } catch (error) {
    if (error.code === '23505') {
      throw new _errors.UniqueViolationError(error.message);
    }

    throw error;
  }
};

/**
 * Makes a query and expects exactly one result.
 *
 * @throws NotFoundError If query returns no rows.
 * @throws DataIntegrityError If query returns multiple rows.
 */
const one = exports.one = async (connection, clientConfiguration, rawSql, values) => {
  const {
    rows
  } = await query(connection, rawSql, values);

  if (rows.length === 0) {
    const ConfigurableNotFoundError = clientConfiguration.errors && clientConfiguration.errors.NotFoundError ? clientConfiguration.errors.NotFoundError : _errors.NotFoundError;

    throw new ConfigurableNotFoundError();
  }

  if (rows.length > 1) {
    throw new _errors.DataIntegrityError();
  }

  return rows[0];
};

/**
 * Makes a query and expects exactly one result.
 *
 * @throws DataIntegrityError If query returns multiple rows.
 */
const maybeOne = exports.maybeOne = async (connection, clientConfiguration, rawSql, values) => {
  const {
    rows
  } = await query(connection, rawSql, values);

  if (rows.length === 0) {
    return null;
  }

  if (rows.length > 1) {
    throw new _errors.DataIntegrityError();
  }

  return rows[0];
};

/**
 * Makes a query and expects exactly one result.
 * Returns value of the first column.
 *
 * @throws NotFoundError If query returns no rows.
 * @throws DataIntegrityError If query returns multiple rows.
 */
const oneFirst = exports.oneFirst = async (connection, clientConfiguration, rawSql, values) => {
  const row = await one(connection, clientConfiguration, rawSql, values);

  const keys = Object.keys(row);

  if (keys.length !== 1) {
    throw new _errors.DataIntegrityError();
  }

  return row[keys[0]];
};

/**
 * Makes a query and expects exactly one result.
 * Returns value of the first column.
 *
 * @throws DataIntegrityError If query returns multiple rows.
 */
const maybeOneFirst = exports.maybeOneFirst = async (connection, clientConfiguration, rawSql, values) => {
  const row = await maybeOne(connection, clientConfiguration, rawSql, values);

  if (!row) {
    return null;
  }

  const keys = Object.keys(row);

  if (keys.length !== 1) {
    throw new _errors.DataIntegrityError();
  }

  return row[keys[0]];
};

/**
 * Makes a query and expects at least 1 result.
 *
 * @throws NotFoundError If query returns no rows.
 */
const many = exports.many = async (connection, clientConfiguration, rawSql, values) => {
  const {
    rows
  } = await query(connection, rawSql, values);

  if (rows.length === 0) {
    const ConfigurableNotFoundError = clientConfiguration.errors && clientConfiguration.errors.NotFoundError ? clientConfiguration.errors.NotFoundError : _errors.NotFoundError;

    throw new ConfigurableNotFoundError();
  }

  return rows;
};

const manyFirst = exports.manyFirst = async (connection, clientConfigurationType, rawSql, values) => {
  const rows = await many(connection, clientConfigurationType, rawSql, values);

  if (rows.length === 0) {
    throw new _errors.DataIntegrityError();
  }

  const keys = Object.keys(rows[0]);

  if (keys.length !== 1) {
    throw new _errors.DataIntegrityError();
  }

  const firstColumnName = keys[0];

  if (typeof firstColumnName !== 'string') {
    throw new _errors.DataIntegrityError();
  }

  return rows.map(row => {
    return row[firstColumnName];
  });
};

/**
 * Makes a query and expects any number of results.
 */
const any = exports.any = async (connection, clientConfiguration, rawSql, values) => {
  const {
    rows
  } = await query(connection, rawSql, values);

  return rows;
};

const anyFirst = exports.anyFirst = async (connection, clientConfigurationType, rawSql, values) => {
  const rows = await any(connection, clientConfigurationType, rawSql, values);

  if (rows.length === 0) {
    return [];
  }

  const keys = Object.keys(rows[0]);

  if (keys.length !== 1) {
    throw new _errors.DataIntegrityError();
  }

  const firstColumnName = keys[0];

  if (typeof firstColumnName !== 'string') {
    throw new _errors.DataIntegrityError();
  }

  return rows.map(row => {
    return row[firstColumnName];
  });
};

const transaction = exports.transaction = async (connection, handler) => {
  await connection.query('START TRANSACTION');

  try {
    const result = await handler(connection);

    await connection.query('COMMIT');

    return result;
  } catch (error) {
    await connection.query('ROLLBACK');

    throw error;
  }
};

const sql = (parts, ...values) => {
  return {
    sql: parts.join('?'),
    values
  };
};

const createConnection = async (connectionConfiguration, clientConfiguration = {}) => {
  const pool = new _pg2.default.Pool(typeof connectionConfiguration === 'string' ? (0, _pgConnectionString.parse)(connectionConfiguration) : connectionConfiguration);

  const connection = await pool.connect();

  let ended = false;

  const bindConnection = {
    any: (0, _utilities.mapTaggedTemplateLiteralInvocation)(any.bind(null, connection, clientConfiguration)),
    anyFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(anyFirst.bind(null, connection, clientConfiguration)),
    end: async () => {
      if (ended) {
        return ended;
      }

      await connection.release();

      ended = pool.end();

      return ended;
    },
    many: (0, _utilities.mapTaggedTemplateLiteralInvocation)(many.bind(null, connection, clientConfiguration)),
    manyFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(manyFirst.bind(null, connection, clientConfiguration)),
    maybeOne: (0, _utilities.mapTaggedTemplateLiteralInvocation)(maybeOne.bind(null, connection, clientConfiguration)),
    maybeOneFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(maybeOneFirst.bind(null, connection, clientConfiguration)),
    one: (0, _utilities.mapTaggedTemplateLiteralInvocation)(one.bind(null, connection, clientConfiguration)),
    oneFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(oneFirst.bind(null, connection, clientConfiguration)),
    query: (0, _utilities.mapTaggedTemplateLiteralInvocation)(query.bind(null, connection)),
    transaction: handler => {
      return transaction(bindConnection, handler);
    }
  };

  return bindConnection;
};

const createPool = (connectionConfiguration, clientConfiguration = {}) => {
  const pool = new _pg2.default.Pool(typeof connectionConfiguration === 'string' ? (0, _pgConnectionString.parse)(connectionConfiguration) : connectionConfiguration);

  const connect = async () => {
    const connection = await pool.connect();

    const bindConnection = {
      any: (0, _utilities.mapTaggedTemplateLiteralInvocation)(any.bind(null, connection, clientConfiguration)),
      anyFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(anyFirst.bind(null, connection, clientConfiguration)),
      many: (0, _utilities.mapTaggedTemplateLiteralInvocation)(many.bind(null, connection, clientConfiguration)),
      manyFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(manyFirst.bind(null, connection, clientConfiguration)),
      maybeOne: (0, _utilities.mapTaggedTemplateLiteralInvocation)(maybeOne.bind(null, connection, clientConfiguration)),
      maybeOneFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(maybeOneFirst.bind(null, connection, clientConfiguration)),
      one: (0, _utilities.mapTaggedTemplateLiteralInvocation)(one.bind(null, connection, clientConfiguration)),
      oneFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(oneFirst.bind(null, connection, clientConfiguration)),
      query: (0, _utilities.mapTaggedTemplateLiteralInvocation)(query.bind(null, connection)),
      release: connection.release.bind(connection),
      transaction: handler => {
        return transaction(bindConnection, handler);
      }
    };

    return bindConnection;
  };

  return {
    any: (0, _utilities.mapTaggedTemplateLiteralInvocation)(any.bind(null, pool, clientConfiguration)),
    anyFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(anyFirst.bind(null, pool, clientConfiguration)),
    connect,
    many: (0, _utilities.mapTaggedTemplateLiteralInvocation)(many.bind(null, pool, clientConfiguration)),
    manyFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(manyFirst.bind(null, pool, clientConfiguration)),
    maybeOne: (0, _utilities.mapTaggedTemplateLiteralInvocation)(maybeOne.bind(null, pool, clientConfiguration)),
    maybeOneFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(maybeOneFirst.bind(null, pool, clientConfiguration)),
    one: (0, _utilities.mapTaggedTemplateLiteralInvocation)(one.bind(null, pool, clientConfiguration)),
    oneFirst: (0, _utilities.mapTaggedTemplateLiteralInvocation)(oneFirst.bind(null, pool, clientConfiguration)),
    query: (0, _utilities.mapTaggedTemplateLiteralInvocation)(query.bind(null, pool)),
    transaction: async handler => {
      log.debug('allocating a new connection to execute the transaction');

      const connection = await connect();

      let result;

      try {
        result = await connection.transaction(handler);
      } finally {
        log.debug('releasing the connection that was earlier secured to execute the transaction');

        await connection.release();
      }

      return result;
    }
  };
};

exports.createConnection = createConnection;
exports.createPool = createPool;
exports.sql = sql;
//# sourceMappingURL=index.js.map